/*
 * xlpmonitor - printer port monitor for Linux/intel boxes
 * (Version 0.2, 2001-05-09)
 *
 * Copyright (C) 2001 by Tom Sato <VEF00200@nifty.ne.jp>
 * http://member.nifty.ne.jp/tsato/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#define PROGRAM_NAME_WITH_VERSION "xlpmonitor (v0.2)"

/*
 * This program is strongly depends to the Linux/intel system.
 *
 * - Status of the printer port will be read directly from the I/O
 *   port of the system.
 *   I/O port will be read via /dev/port pseudo device.
 *
 *   To read /dev/port, this program should chgrp'ed to kmem group
 *   and setgid'ed.
 *
 * - When the port address was not explicitly specified, the address
 *   will be obtained from the boot message, by seeking the message
 *   like ``parport0: PC-style at 0x378 [SPP,ECP,ECPEPP,ECPPS2]''
 *   printed by the parport module.
 *
 * Following values are to interpret the value read from the I/O port
 * (base address + 1), and was extracted from /usr/include/linux/lp.h.
 */
#define LP_PBUSY        0x80  /* inverted input, active high */
#define LP_PACK         0x40  /* unchanged input, active low */
#define LP_POUTPA       0x20  /* unchanged input, active high */
#define LP_PSELECD      0x10  /* unchanged input, active high */
#define LP_PERRORP      0x08  /* unchanged input, active low */

/*
 * Resources and options
 */
struct appres_struct {
  Boolean debug;
  Boolean print;
  int address;
  int default_address;
  int port;
  String port_device;
  String log_message_file;

  int interval;
  int quick_interval;

  XFontStruct *font;
  Pixel normal_color;
  Pixel active_color;
  Pixel inactive_color;
  Pixel error_color;
} appres;

#define Offset(entry) XtOffset(struct appres_struct *, entry)
static XtResource application_resources[] = {
  { "debug", "Debug", XtRBoolean, sizeof(Boolean),
    Offset(debug), XtRImmediate, (XtPointer)FALSE },
  { "print", "Print", XtRBoolean, sizeof(Boolean),
    Offset(print), XtRImmediate, (XtPointer)FALSE },
  { "address", "Address", XtRInt, sizeof(int),
    Offset(address), XtRImmediate, (XtPointer)-1 },
  { "defaultAddress", "Address", XtRInt, sizeof(int),
    Offset(default_address), XtRImmediate, (XtPointer)0x0378 },
  { "port", "Port", XtRInt, sizeof(int),
    Offset(port), XtRImmediate, (XtPointer)0 },
  { "portDevice", "PortDevice", XtRString, sizeof(char *),
    Offset(port_device), XtRImmediate, "/dev/port" },
  { "logMessageFile", "LogMessageFile", XtRString, sizeof(char *),
    Offset(log_message_file), XtRImmediate, "/var/log/messages" },

  { "interval", "Interval", XtRInt, sizeof(int),
    Offset(interval), XtRImmediate, (XtPointer)100 },
  { "quickInterval", "Interval", XtRInt, sizeof(int),
    Offset(quick_interval), XtRImmediate, (XtPointer)10 },

  { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    Offset(font), XtRString, XtDefaultFont },
  { "normalColor", XtCForeground, XtRPixel, sizeof(Pixel),
    Offset(normal_color), XtRString, "black"},
  { "activeColor", XtCForeground, XtRPixel, sizeof(Pixel),
    Offset(active_color), XtRString, "green"},
  { "inactiveColor", XtCForeground, XtRPixel, sizeof(Pixel),
    Offset(inactive_color), XtRString, "gray60"},
  { "errorColor", XtCForeground, XtRPixel, sizeof(Pixel),
    Offset(error_color), XtRString, "red"},
};
#undef Offset

static XrmOptionDescRec options[] = {
  { "-debug", ".debug", XrmoptionNoArg, "True" },
  { "-print", ".print", XrmoptionNoArg, "True" },
  { "-address", ".address", XrmoptionSepArg, NULL },
  { "-port", ".port", XrmoptionSepArg, NULL },
  { "-interval", ".interval", XrmoptionSepArg, NULL },
};

/*
 * Global variables
 */
static Widget toplevel;
static Widget canvas;
static Atom wm_delete_window = None;

static GC gc;
static int font_height;

static enum LEDs { POWER_LED, READY_LED, ONLINE_LED, PAPER_LED } dummy;
static const char *led_label[] = { "POWER", "READY", "ONLINE", "PAPER" };

static Boolean canvas_exposed = True;

static uid_t euid;
static gid_t egid;

/*
 * Draw the indicator with the specified color
 */
static void DrawIndicator(enum LEDs led, Pixel color)
{
  Display *dpy = XtDisplay(canvas);
  Window win = XtWindow(canvas);

  int ind_ht = font_height * 0.8;
  int ind_wd = ind_ht * 1.5;
  int x1 = ind_wd * 0.3;
  int y;
  const char *label;

  if (appres.print) return;

  y = font_height * (led + 1.5);
  XSetForeground(dpy, gc, color);
  XFillRectangle(dpy, win, gc, x1, y, ind_wd, ind_ht);
  XSetForeground(dpy, gc, appres.inactive_color);
  XDrawRectangle(dpy, win, gc, x1, y, ind_wd, ind_ht);

  label = led_label[led];
  XSetForeground(dpy, gc, appres.normal_color);
  XDrawString(dpy, win, gc, x1 + ind_wd * 1.3, y + ind_ht, label, strlen(label));

  if (led == POWER_LED) {
    label = "- Printer -";
    XDrawString(dpy, win, gc, x1, font_height, label, strlen(label));
  }
}

/*
 * Read status of the printer port and redraw the canvas.
 * Short message indicating the status will be returned.
 */
static const char *Redraw(XtPointer closure, XtIntervalId *id)
{
  static Boolean first = True;
  static int dev_port_fd;
  static unsigned char last_state = 0;
  static const char *msg = "unknown";

  Display *dpy = XtDisplay(canvas);
  Window win = XtWindow(canvas);
  int interval;
  unsigned char state;
  int busy, ack, pe, sel, err;

  if (first) {
    first = False;

    gc = XCreateGC(dpy, win, 0, NULL);
    XSetFont(dpy, gc, appres.font->fid);
    font_height = appres.font->max_bounds.ascent + appres.font->max_bounds.descent;

    if (dev_port_fd == 0) {
      seteuid(euid);
      setegid(egid);
      dev_port_fd = open(appres.port_device, O_RDONLY);
      if (dev_port_fd < 0) {
	fprintf(stderr,
		"xlpmonitor: Can't monitor the printer port: can't open %s: %s\n",
		appres.port_device, strerror(errno));
      }
      seteuid(getuid());
      setegid(getgid());
    }
  }

  interval = appres.interval;

  if (dev_port_fd < 0) {
    msg = "cannot check";
    if (canvas_exposed) {
      DrawIndicator(POWER_LED , appres.error_color);
      DrawIndicator(PAPER_LED , appres.error_color);
      DrawIndicator(ONLINE_LED, appres.error_color);
      DrawIndicator(READY_LED , appres.error_color);
    }
  } else {
    lseek(dev_port_fd, appres.address + 1, SEEK_SET);
    read(dev_port_fd, &state, 1);

    if (canvas_exposed || state != last_state) {
      canvas_exposed = False;
      last_state = state;

      busy = (state & LP_PBUSY)   ? 0 : 1;
      ack  = (state & LP_PACK)    ? 0 : 1;
      pe   = (state & LP_POUTPA)  ? 1 : 0;
      sel  = (state & LP_PSELECD) ? 1 : 0;
      err  = (state & LP_PERRORP) ? 0 : 1;

      if (appres.debug) printf("BASE=0x%x BUSY=%d ACK=%d PE=%d SEL=%d ERR=%d\n",
			       appres.address, busy, ack, pe, sel, err);

      if (!sel) {  /* printer deselected */
	msg = "deselect";
	DrawIndicator(POWER_LED , appres.error_color);
	DrawIndicator(PAPER_LED , appres.inactive_color);
	DrawIndicator(ONLINE_LED, appres.inactive_color);
      } else if (!err && pe && busy && !ack) {  /* printer disconnected */
	msg = "disconnected";
	DrawIndicator(POWER_LED , appres.inactive_color);
	DrawIndicator(PAPER_LED , appres.inactive_color);
	DrawIndicator(ONLINE_LED, appres.inactive_color);
      } else if (!err && pe && busy && ack) {  /* power-down */
	msg = "power down";
	DrawIndicator(POWER_LED , appres.inactive_color);
	DrawIndicator(PAPER_LED , appres.inactive_color);
	DrawIndicator(ONLINE_LED, appres.inactive_color);
      } else if (pe) {  /* paper empty */
	msg = "paper empty";
	DrawIndicator(POWER_LED , appres.active_color);
	DrawIndicator(PAPER_LED , appres.error_color);
	DrawIndicator(ONLINE_LED, appres.inactive_color);
      } else if (err) {  /* error detected, or printer offline */
	msg = "error detected, or printer offline";
	DrawIndicator(POWER_LED , appres.active_color);
	DrawIndicator(PAPER_LED , appres.inactive_color);
	DrawIndicator(ONLINE_LED, appres.inactive_color);
      } else {  /* ready to use */
	msg = "ready";
	interval = appres.quick_interval;
	DrawIndicator(POWER_LED , appres.active_color);
	DrawIndicator(PAPER_LED , appres.active_color);
	DrawIndicator(ONLINE_LED, appres.active_color);
      }
      DrawIndicator(READY_LED, busy ? appres.inactive_color : appres.active_color);
    }
  }

  if (!appres.print)
    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel),
		    interval, (XtTimerCallbackProc)Redraw, NULL);

  return msg;
}

/*
 * When it is exposed, we must redraw the canvas even if state unchanged.
 */
static void ExposeProc(Widget w, XEvent *event,
		       String *pars, Cardinal *n_pars)
{
  canvas_exposed = True;
}

/*
 * WM_DELETE_WINDOW has been sent - terminate the program.
 */
static void DeleteWindowProc(Widget w, XEvent *event,
                             String *pars, Cardinal *n_pars)
{
  XtDestroyApplicationContext(XtWidgetToApplicationContext(toplevel));
  exit(0);
}

/*
 * The main routine:
 * Usally, a window to display the status of the printer will be created.
 * If invoked with -print option, print a short message describing
 * the status of the printer and terminate immediately.
 */
int main(int argc, char *argv[])
{
  static XtActionsRec actions[] = {
    { "DeleteWindowProc", DeleteWindowProc },
    { "ExposeProc", ExposeProc },
  };
  static String fallback_resources[] = {
    "*title: LP Monitor",
    "*iconName: LP Monitor",
    "*Font: -*-lucida-medium-r-normal-*-12-*-75-75-*-*-*-*",
    "*background: gray85",
    "*foreground: black",
    "*canvas.width: 75",
    "*canvas.height: 75",
    "*translations: #override \
	<Message>WM_PROTOCOLS: DeleteWindowProc()\\n\
	<Expose>: ExposeProc()",
    NULL,
  };

  FILE *fp;
  char str[200], *cp;
  char key[50];
  int n;
  int phase;
  char filename[100];

  euid = geteuid();
  egid = getegid();
  seteuid(getuid());
  setegid(getgid());
  
  toplevel = XtVaAppInitialize(NULL, "Xlpmonitor",
                               options, XtNumber(options),
                               &argc, argv, fallback_resources, NULL);
  XtGetApplicationResources(toplevel, &appres,
            application_resources, XtNumber(application_resources),
            NULL, 0);
  XtAppAddActions(XtWidgetToApplicationContext(toplevel), actions, XtNumber(actions));

  canvas = XtVaCreateManagedWidget("canvas", coreWidgetClass, toplevel, NULL);

  if (appres.debug) {
    fprintf(stderr, PROGRAM_NAME_WITH_VERSION
	    "printer port monitor for Linux/intel boxes\n");
    fprintf(stderr, "Copyright (C) 2001 by Tom Sato <VEF00200@nifty.ne.jp>\n\n");
    fprintf(stderr, "This program is free software with ABSOLUTELY NO WARRANTY,\n");
    fprintf(stderr, "distributed under the terms of the GNU General Public License.\n\n");
  }

  if (appres.address < 0) {
    sprintf(filename, "/proc/parport/%d/hardware", appres.port);
    fp = fopen(filename, "r");
    if (fp == NULL) {
      fprintf(stderr, "xlpmonitor: Can't read %s: %s\n", filename, strerror(errno));
      fprintf(stderr, "xlpmonitor: \"parport\" module not loaded?\n");
    } else {
      while (fgets(str, sizeof(str) - 1, fp)) {
	if (sscanf(str, "base: %x", &appres.address) == 1) break;
      }
      fclose(fp);
    }
  }
  if (appres.address < 0) {
    fprintf(stderr, "xlpmonitor: Can't find base address in %s - scanning %s\n",
	    filename, "/proc/ioports");
    /* check /proc/ioports for line like ``0378-037f: parport0'' */
    fp = fopen("/proc/ioports", "r");
    if (fp != NULL) {
      sprintf(key, "parport%d\n", appres.port);
      while (fgets(str, sizeof(str) - 1, fp)) {
	if (strstr(str, key) != NULL
	    || (appres.port == 0 && strstr(str, "lp\n") != NULL)) {
	  sscanf(str, "%x", &appres.address);
	  break;
	}
      }
      fclose(fp);
    }
  }
  if (appres.address < 0 && 0 < strlen(appres.log_message_file)) {
    fprintf(stderr, "xlpmonitor: Can't find base address - scanning %s\n",
	    appres.log_message_file);

    seteuid(euid);
    setegid(egid);
    fp = fopen(appres.log_message_file, "r");
    seteuid(getuid());
    setegid(getgid());

    if (fp == NULL) {
      fprintf(stderr, "xlpmonitor: Can't open %s: %s\n",
	      appres.log_message_file, strerror(errno));
    } else {
      while (fgets(str, sizeof(str) - 1, fp)) {
	/* find port address of the parallel port by scanning the boot message
	   for line like ``parport0: PC-style at 0x378 [SPP,ECP,ECPEPP,ECPPS2]'' */
	if ((cp = strstr(str, "parport")) != NULL
	    && sscanf(cp, "parport%d", &n) == 1
	    && n == appres.port) {
	  cp = strstr(cp, "0x");
	  if (cp != NULL) {
	    sscanf(cp, "%x", &appres.address);
	    break;
	  }
	}
      }
      fclose(fp);
    }
  }
  
  if (appres.address < 0) {
    fprintf(stderr, "xlpmonitor: Can't find base address - using default (0x%04x)\n",
	    appres.default_address);
    fprintf(stderr, "xlpmonitor: You can set the address with \"-address\" option\n");
    appres.address = appres.default_address;
  }
  if (appres.debug)
    fprintf(stderr, "xlpmonitor: Port address is 0x%04x\n", appres.address);

  if (appres.print) {
    printf("%s\n", Redraw(NULL, NULL));
  } else {
    XtRealizeWidget(toplevel);
    wm_delete_window = XInternAtom(XtDisplay(toplevel), "WM_DELETE_WINDOW", FALSE);
    XSetWMProtocols(XtDisplay(toplevel), XtWindow(toplevel), &wm_delete_window, 1);

    Redraw(NULL, NULL);

    XtAppMainLoop(XtWidgetToApplicationContext(toplevel));
  }
}
